"""Main class ActionRules."""

import itertools
from collections import defaultdict
from typing import Optional

import pandas as pd

from .candidates.candidate_generator import CandidateGenerator
from .output.output import Output
from .rules.rules import Rules


class ActionRules:
    """
    A class used to generate action rules for a given dataset.

    Attributes
    ----------
    min_stable_attributes : int
        The minimum number of stable attributes required.
    min_flexible_attributes : int
        The minimum number of flexible attributes required.
    min_undesired_support : int
        The minimum support for the undesired state.
    min_undesired_confidence : float
        The minimum confidence for the undesired state.
    min_desired_support : int
        The minimum support for the desired state.
    min_desired_confidence : float
        The minimum confidence for the desired state.
    verbose : bool, optional
        If True, enables verbose output.
    rules : Rules, optional
        Stores the generated rules.
    output : Output, optional
        Stores the generated action rules.

    Methods
    -------
    fit(data, stable_attributes, flexible_attributes, target, undesired_state, desired_state)
        Generates action rules based on the provided dataset and parameters.
    get_bindings(data, stable_attributes, flexible_attributes, target)
        Binds attributes to corresponding columns in the dataset.
    get_stop_list(stable_items_binding, flexible_items_binding)
        Generates a stop list to prevent certain combinations of attributes.
    get_split_tables(data, target_items_binding, target)
        Splits the dataset into tables based on target item bindings.
    get_rules()
        Returns the generated action rules if available.
    """

    def __init__(
        self,
        min_stable_attributes: int,
        min_flexible_attributes: int,
        min_undesired_support: int,
        min_undesired_confidence: float,
        min_desired_support: int,
        min_desired_confidence: float,
        verbose=False,
    ):
        """
        Initialize the ActionRules class with the specified parameters.

        Parameters
        ----------
        min_stable_attributes : int
            The minimum number of stable attributes required.
        min_flexible_attributes : int
            The minimum number of flexible attributes required.
        min_undesired_support : int
            The minimum support for the undesired state.
        min_undesired_confidence : float
            The minimum confidence for the undesired state.
        min_desired_support : int
            The minimum support for the desired state.
        min_desired_confidence : float
            The minimum confidence for the desired state.
        verbose : bool, optional
            If True, enables verbose output. Default is False.
        """
        self.min_stable_attributes = min_stable_attributes
        self.min_flexible_attributes = min_flexible_attributes
        self.min_undesired_support = min_undesired_support
        self.min_desired_support = min_desired_support
        self.min_undesired_confidence = min_undesired_confidence
        self.min_desired_confidence = min_desired_confidence
        self.verbose = verbose
        self.rules = None  # type: Optional[Rules]
        self.output = None  # type: Optional[Output]

    def fit(
        self,
        data: pd.DataFrame,
        stable_attributes: list,
        flexible_attributes: list,
        target: str,
        undesired_state: str,
        desired_state: str,
    ):
        """
        Generate action rules based on the provided dataset and parameters.

        Parameters
        ----------
        data : pd.DataFrame
            The dataset to generate action rules from.
        stable_attributes : list
            List of stable attributes.
        flexible_attributes : list
            List of flexible attributes.
        target : str
            The target attribute.
        undesired_state : str
            The undesired state of the target attribute.
        desired_state : str
            The desired state of the target attribute.
        """
        data = data.astype(str)
        data_stable = pd.get_dummies(data[stable_attributes], sparse=False, prefix_sep='_<item_stable>_')
        data_flexible = pd.get_dummies(data[flexible_attributes], sparse=False, prefix_sep='_<item_flexible>_')
        data_target = pd.get_dummies(data[[target]], sparse=False, prefix_sep='_<item_target>_')
        data = pd.concat([data_stable, data_flexible, data_target], axis=1)
        stable_items_binding, flexible_items_binding, target_items_binding = self.get_bindings(
            data, stable_attributes, flexible_attributes, target
        )
        stop_list = self.get_stop_list(stable_items_binding, flexible_items_binding)
        frames = self.get_split_tables(data, target_items_binding, target)
        undesired_state = target + '_<item_target>_' + str(undesired_state)
        desired_state = target + '_<item_target>_' + str(desired_state)

        stop_list_itemset = []  # type: list

        candidates_queue = [
            {
                'ar_prefix': tuple(),
                'itemset_prefix': tuple(),
                'stable_items_binding': stable_items_binding,
                'flexible_items_binding': flexible_items_binding,
                'undesired_mask': None,
                'desired_mask': None,
                'actionable_attributes': 0,
            }
        ]
        k = 0
        self.rules = Rules(undesired_state, desired_state)
        candidate_generator = CandidateGenerator(
            frames,
            self.min_stable_attributes,
            self.min_flexible_attributes,
            self.min_undesired_support,
            self.min_desired_support,
            self.min_undesired_confidence,
            self.min_desired_confidence,
            undesired_state,
            desired_state,
            self.rules,
        )
        while len(candidates_queue) > 0:
            candidate = candidates_queue.pop(0)
            if len(candidate['ar_prefix']) > k:
                k += 1
                self.rules.prune_classification_rules(k, stop_list)
            new_candidates = candidate_generator.generate_candidates(
                **candidate,
                stop_list=stop_list,
                stop_list_itemset=stop_list_itemset,
                undesired_state=undesired_state,
                desired_state=desired_state,
                verbose=self.verbose,
            )
            candidates_queue += new_candidates
        self.rules.generate_action_rules()
        self.output = Output(self.rules.action_rules, target)

    def get_bindings(
        self, data: pd.DataFrame, stable_attributes: list, flexible_attributes: list, target: str
    ) -> tuple:
        """
        Bind attributes to corresponding columns in the dataset.

        Parameters
        ----------
        data : pd.DataFrame
            The dataset containing the attributes.
        stable_attributes : list
            List of stable attributes.
        flexible_attributes : list
            List of flexible attributes.
        target : str
            The target attribute.

        Returns
        -------
        tuple
            A tuple containing the bindings for stable attributes, flexible attributes, and target items.
        """
        stable_items_binding = defaultdict(lambda: [])
        flexible_items_binding = defaultdict(lambda: [])
        target_items_binding = defaultdict(lambda: [])

        for col in data.columns:
            is_continue = False
            # stable
            for attribute in stable_attributes:
                if col.startswith(attribute + '_<item_stable>_'):
                    stable_items_binding[attribute].append(col)
                    is_continue = True
                    break
            if is_continue is True:
                continue
            # flexible
            for attribute in flexible_attributes:
                if col.startswith(attribute + '_<item_flexible>_'):
                    flexible_items_binding[attribute].append(col)
                    is_continue = True
                    break
            if is_continue is True:
                continue
            # target
            if col.startswith(target + '_<item_target>_'):
                target_items_binding[target].append(col)
        return stable_items_binding, flexible_items_binding, target_items_binding

    def get_stop_list(self, stable_items_binding: dict, flexible_items_binding: dict) -> list:
        """
        Generate a stop list to prevent certain combinations of attributes.

        Parameters
        ----------
        stable_items_binding : dict
            Dictionary containing bindings for stable items.
        flexible_items_binding : dict
            Dictionary containing bindings for flexible items.

        Returns
        -------
        list
            A list of stop combinations.
        """
        stop_list = []
        for items in stable_items_binding.values():
            for stop_couple in itertools.product(items, repeat=2):
                stop_list.append(tuple(stop_couple))
        for item in flexible_items_binding.keys():
            stop_list.append(tuple([item, item]))
        return stop_list

    def get_split_tables(self, data: pd.DataFrame, target_items_binding: dict, target: str) -> dict:
        """
        Split the dataset into tables based on target item bindings.

        Parameters
        ----------
        data : pd.DataFrame
            The dataset to be split.
        target_items_binding : dict
            Dictionary containing bindings for target items.
        target : str
            The target attribute.

        Returns
        -------
        dict
            A dictionary containing the split tables.
        """
        frames = {}
        for item in target_items_binding[target]:
            mask = data[item] == 1
            frames[item] = data[mask]
        return frames

    def get_rules(self) -> Optional[Output]:
        """
        Return the generated action rules if available.

        Returns
        -------
        Optional[Output]
            The generated action rules, or None if no rules have been generated.
        """
        if self.output is None:
            return None
        else:
            return self.output



#!/usr/bin/env python
"""Tests for `action_rules` package."""

import pandas as pd
import pytest

from action_rules.candidates.candidate_generator import CandidateGenerator
from action_rules.rules.rules import Rules


@pytest.fixture
def sample_frames():
    """Fixture for sample frames to be used in tests."""
    frames = {
        'status_<item_target>_default': pd.DataFrame(
            {
                'age_<item_stable>_30': [1, 0, 0],
                'income_<item_flexible>_low': [1, 0, 0],
                'status_<item_target>_default': [1, 0, 0],
            }
        ),
        'status_<item_target>_paid': pd.DataFrame(
            {
                'age_<item_stable>_30': [0, 1, 0],
                'income_<item_flexible>_low': [0, 1, 0],
                'status_<item_target>_paid': [0, 1, 0],
            }
        ),
    }
    return frames


@pytest.fixture
def rules():
    """Fixture for Rules instance."""
    return Rules('status_<item_target>_default', 'status_<item_target>_paid')


@pytest.fixture
def candidate_generator(sample_frames, rules):
    """Fixture for CandidateGenerator instance."""
    return CandidateGenerator(
        frames=sample_frames,
        min_stable_attributes=1,
        min_flexible_attributes=1,
        min_undesired_support=1,
        min_desired_support=1,
        min_undesired_confidence=0.5,
        min_desired_confidence=0.5,
        undesired_state='status_<item_target>_default',
        desired_state='status_<item_target>_paid',
        rules=rules,
    )


def test_generate_candidates(candidate_generator):
    """Test the generate_candidates method of CandidateGenerator."""
    ar_prefix = tuple()
    itemset_prefix = tuple()
    stable_items_binding = {'age': ['age_<item_stable>_30']}
    flexible_items_binding = {'income': ['income_<item_flexible>_low']}
    undesired_mask = pd.Series([1, 0, 0])
    desired_mask = pd.Series([0, 1, 0])
    actionable_attributes = 0
    stop_list = []
    stop_list_itemset = []
    undesired_state = 'status_<item_target>_default'
    desired_state = 'status_<item_target>_paid'
    new_branches = candidate_generator.generate_candidates(
        ar_prefix,
        itemset_prefix,
        stable_items_binding,
        flexible_items_binding,
        undesired_mask,
        desired_mask,
        actionable_attributes,
        stop_list,
        stop_list_itemset,
        undesired_state,
        desired_state,
        verbose=True,
    )
    assert len(new_branches) > 0


def test_get_frames(candidate_generator):
    """Test the get_frames method of CandidateGenerator."""
    undesired_mask = pd.Series([1, 0, 0])
    desired_mask = pd.Series([0, 1, 0])
    undesired_state = 'status_<item_target>_default'
    desired_state = 'status_<item_target>_paid'
    undesired_frame, desired_frame = candidate_generator.get_frames(
        undesired_mask, desired_mask, undesired_state, desired_state
    )
    assert not undesired_frame.empty
    assert not desired_frame.empty


def test_reduce_candidates_by_min_attributes(candidate_generator):
    """Test the reduce_candidates_by_min_attributes method of CandidateGenerator."""
    stable_items_binding = {'age': ['age_<item_stable>_30', 'age_<item_stable>_40', 'age_<item_stable>_50']}
    flexible_items_binding = {'income': ['income_<item_flexible>_low', 'income_<item_flexible>_high']}
    k = 1
    actionable_attributes = 0
    reduced_stable_items_binding, reduced_flexible_items_binding = (
        candidate_generator.reduce_candidates_by_min_attributes(
            k, actionable_attributes, stable_items_binding, flexible_items_binding
        )
    )
    assert len(reduced_stable_items_binding) == 1
    assert len(reduced_flexible_items_binding) == 0
    k = 2
    actionable_attributes = 0
    reduced_stable_items_binding, reduced_flexible_items_binding = (
        candidate_generator.reduce_candidates_by_min_attributes(
            k, actionable_attributes, stable_items_binding, flexible_items_binding
        )
    )
    assert len(reduced_stable_items_binding) == 1
    assert len(reduced_flexible_items_binding) == 1


def test_process_stable_candidates(candidate_generator):
    """Test the process_stable_candidates method of CandidateGenerator."""
    ar_prefix = tuple()
    itemset_prefix = tuple()
    reduced_stable_items_binding = {'age': ['age_<item_stable>_30']}
    stop_list = []
    stable_candidates = {'age': ['age_<item_stable>_30']}
    undesired_frame = pd.DataFrame({'age_<item_stable>_30': [1, 0, 0]})
    desired_frame = pd.DataFrame({'age_<item_stable>_30': [0, 1, 0]})
    new_branches = []
    candidate_generator.process_stable_candidates(
        ar_prefix,
        itemset_prefix,
        reduced_stable_items_binding,
        stop_list,
        stable_candidates,
        undesired_frame,
        desired_frame,
        new_branches,
        verbose=True,
    )
    assert len(new_branches) > 0


def test_process_flexible_candidates(candidate_generator):
    """Test the process_flexible_candidates method of CandidateGenerator."""
    ar_prefix = tuple()
    itemset_prefix = tuple()
    reduced_flexible_items_binding = {'income': ['income_<item_flexible>_low']}
    stop_list = []
    stop_list_itemset = []
    flexible_candidates = {'income': ['income_<item_flexible>_low']}
    undesired_frame = pd.DataFrame({'income_<item_flexible>_low': [1, 0, 0]})
    desired_frame = pd.DataFrame({'income_<item_flexible>_low': [0, 1, 0]})
    actionable_attributes = 0
    new_branches = []
    candidate_generator.process_flexible_candidates(
        ar_prefix,
        itemset_prefix,
        reduced_flexible_items_binding,
        stop_list,
        stop_list_itemset,
        flexible_candidates,
        undesired_frame,
        desired_frame,
        actionable_attributes,
        new_branches,
        verbose=True,
    )
    assert len(new_branches) > 0


def test_process_items(candidate_generator):
    """Test the process_items method of CandidateGenerator."""
    attribute = 'income'
    items = ['income_<item_flexible>_low']
    itemset_prefix = tuple()
    stop_list_itemset = []
    undesired_frame = pd.DataFrame({'income_<item_flexible>_low': [1, 0, 0]})
    desired_frame = pd.DataFrame({'income_<item_flexible>_low': [0, 1, 0]})
    flexible_candidates = {'income': ['income_<item_flexible>_low']}
    undesired_states, desired_states, undesired_count, desired_count = candidate_generator.process_items(
        attribute,
        items,
        itemset_prefix,
        stop_list_itemset,
        undesired_frame,
        desired_frame,
        flexible_candidates,
        verbose=True,
    )
    assert len(undesired_states) > 0
    assert len(desired_states) > 0


def test_update_new_branches(candidate_generator):
    """Test the update_new_branches method of CandidateGenerator."""
    new_branches = [{'item': 'income_<item_flexible>_low'}]
    stable_candidates = {'age': ['age_<item_stable>_30']}
    flexible_candidates = {'income': ['income_<item_flexible>_low']}
    candidate_generator.update_new_branches(new_branches, stable_candidates, flexible_candidates)
    assert 'stable_items_binding' in new_branches[0]
    assert 'flexible_items_binding' in new_branches[0]



"""Class CandidateGenerator."""

import copy

import pandas as pd

from action_rules.rules import Rules


class CandidateGenerator:
    """
    A class used to generate candidate action rules for a given dataset.

    Attributes
    ----------
    frames : dict
        Dictionary containing data frames for undesired and desired states.
    min_stable_attributes : int
        Minimum number of stable attributes required.
    min_flexible_attributes : int
        Minimum number of flexible attributes required.
    min_undesired_support : int
        Minimum support for the undesired state.
    min_desired_support : int
        Minimum support for the desired state.
    min_undesired_confidence : float
        Minimum confidence for the undesired state.
    min_desired_confidence : float
        Minimum confidence for the desired state.
    undesired_state : str
        The undesired state of the target attribute.
    desired_state : str
        The desired state of the target attribute.
    rules : Rules
        Rules object to store the generated classification rules.

    Methods
    -------
    generate_candidates(ar_prefix, itemset_prefix, stable_items_binding, flexible_items_binding, undesired_mask,
    desired_mask, actionable_attributes, stop_list, stop_list_itemset, undesired_state, desired_state, verbose=False)
        Generate candidate action rules.
    get_frames(undesired_mask, desired_mask, undesired_state, desired_state)
        Get the frames for the undesired and desired states.
    reduce_candidates_by_min_attributes(k, actionable_attributes, stable_items_binding, flexible_items_binding)
        Reduce the candidate sets based on minimum attributes.
    process_stable_candidates(ar_prefix, itemset_prefix, reduced_stable_items_binding, stop_list, stable_candidates,
    undesired_frame, desired_frame, new_branches, verbose)
        Process stable candidates to generate new branches.
    process_flexible_candidates(ar_prefix, itemset_prefix, reduced_flexible_items_binding, stop_list, stop_list_itemset,
    flexible_candidates, undesired_frame, desired_frame, actionable_attributes, new_branches, verbose)
        Process flexible candidates to generate new branches.
    process_items(attribute, items, itemset_prefix, stop_list_itemset, undesired_frame, desired_frame,
    flexible_candidates, verbose)
        Process items to generate states and counts.
    update_new_branches(new_branches, stable_candidates, flexible_candidates)
        Update new branches with stable and flexible candidates.
    in_stop_list(ar_prefix, stop_list)
        Check if the action rule prefix is in the stop list.
    """

    def __init__(
        self,
        frames: dict,
        min_stable_attributes: int,
        min_flexible_attributes: int,
        min_undesired_support: int,
        min_desired_support: int,
        min_undesired_confidence: float,
        min_desired_confidence: float,
        undesired_state: str,
        desired_state: str,
        rules: Rules,
    ):
        """
        Initialize the CandidateGenerator class with the specified parameters.

        Parameters
        ----------
        frames : dict
            Dictionary containing data frames for undesired and desired states.
        min_stable_attributes : int
            Minimum number of stable attributes required.
        min_flexible_attributes : int
            Minimum number of flexible attributes required.
        min_undesired_support : int
            Minimum support for the undesired state.
        min_desired_support : int
            Minimum support for the desired state.
        min_undesired_confidence : float
            Minimum confidence for the undesired state.
        min_desired_confidence : float
            Minimum confidence for the desired state.
        undesired_state : str
            The undesired state of the target attribute.
        desired_state : str
            The desired state of the target attribute.
        rules : Rules
            Rules object to store the generated classification rules.
        """
        self.frames = frames
        self.min_stable_attributes = min_stable_attributes
        self.min_flexible_attributes = min_flexible_attributes
        self.min_undesired_support = min_undesired_support
        self.min_desired_support = min_desired_support
        self.min_undesired_confidence = min_undesired_confidence
        self.min_desired_confidence = min_desired_confidence
        self.undesired_state = undesired_state
        self.desired_state = desired_state
        self.rules = rules

    def generate_candidates(
        self,
        ar_prefix: tuple,
        itemset_prefix: tuple,
        stable_items_binding: dict,
        flexible_items_binding: dict,
        undesired_mask: pd.Series,
        desired_mask: pd.Series,
        actionable_attributes: int,
        stop_list: list,
        stop_list_itemset: list,
        undesired_state: str,
        desired_state: str,
        verbose: bool = False,
    ) -> list:
        """
        Generate candidate action rules.

        Parameters
        ----------
        ar_prefix : tuple
            Prefix of the action rule.
        itemset_prefix : tuple
            Prefix of the itemset.
        stable_items_binding : dict
            Dictionary containing bindings for stable items.
        flexible_items_binding : dict
            Dictionary containing bindings for flexible items.
        undesired_mask : pd.Series
            Mask for the undesired state.
        desired_mask : pd.Series
            Mask for the desired state.
        actionable_attributes : int
            Number of actionable attributes.
        stop_list : list
            List of stop combinations.
        stop_list_itemset : list
            List of stop itemsets.
        undesired_state : str
            The undesired state of the target attribute.
        desired_state : str
            The desired state of the target attribute.
        verbose : bool, optional
            If True, enables verbose output. Default is False.

        Returns
        -------
        list
            List of new branches generated.
        """
        k = len(itemset_prefix) + 1
        reduced_stable_items_binding, reduced_flexible_items_binding = self.reduce_candidates_by_min_attributes(
            k, actionable_attributes, stable_items_binding, flexible_items_binding
        )

        undesired_frame, desired_frame = self.get_frames(undesired_mask, desired_mask, undesired_state, desired_state)
        stable_candidates = copy.deepcopy(stable_items_binding)
        flexible_candidates = copy.deepcopy(flexible_items_binding)

        new_branches = []  # type: list

        self.process_stable_candidates(
            ar_prefix,
            itemset_prefix,
            reduced_stable_items_binding,
            stop_list,
            stable_candidates,
            undesired_frame,
            desired_frame,
            new_branches,
            verbose,
        )
        self.process_flexible_candidates(
            ar_prefix,
            itemset_prefix,
            reduced_flexible_items_binding,
            stop_list,
            stop_list_itemset,
            flexible_candidates,
            undesired_frame,
            desired_frame,
            actionable_attributes,
            new_branches,
            verbose,
        )
        self.update_new_branches(new_branches, stable_candidates, flexible_candidates)

        return new_branches

    def get_frames(
        self, undesired_mask: pd.Series, desired_mask: pd.Series, undesired_state: str, desired_state: str
    ) -> tuple:
        """
        Get the frames for the undesired and desired states.

        Parameters
        ----------
        undesired_mask : pd.Series
            Mask for the undesired state.
        desired_mask : pd.Series
            Mask for the desired state.
        undesired_state : str
            The undesired state of the target attribute.
        desired_state : str
            The desired state of the target attribute.

        Returns
        -------
        tuple
            Tuple containing the frames for the undesired and desired states.
        """
        if undesired_mask is None:
            return self.frames[undesired_state], self.frames[desired_state]
        else:
            undesired_frame = self.frames[undesired_state].multiply(undesired_mask, axis="index")
            desired_frame = self.frames[desired_state].multiply(desired_mask, axis="index")
            return undesired_frame, desired_frame

    def reduce_candidates_by_min_attributes(
        self, k: int, actionable_attributes: int, stable_items_binding: dict, flexible_items_binding: dict
    ) -> tuple:
        """
        Reduce the candidate sets based on minimum attributes.

        Parameters
        ----------
        k : int
            Length of the itemset prefix plus one.
        actionable_attributes : int
            Number of actionable attributes.
        stable_items_binding : dict
            Dictionary containing bindings for stable items.
        flexible_items_binding : dict
            Dictionary containing bindings for flexible items.

        Returns
        -------
        tuple
            Tuple containing the reduced stable and flexible items bindings.
        """
        number_of_stable_attributes = len(stable_items_binding) - (self.min_stable_attributes - k)
        if k > self.min_stable_attributes:
            number_of_flexible_attributes = len(flexible_items_binding) - (
                self.min_flexible_attributes - actionable_attributes - 1
            )
        else:
            number_of_flexible_attributes = 0
        reduced_stable_items_binding = {
            k: stable_items_binding[k] for k in list(stable_items_binding.keys())[:number_of_stable_attributes]
        }
        reduced_flexible_items_binding = {
            k: flexible_items_binding[k] for k in list(flexible_items_binding.keys())[:number_of_flexible_attributes]
        }
        return reduced_stable_items_binding, reduced_flexible_items_binding

    def process_stable_candidates(
        self,
        ar_prefix: tuple,
        itemset_prefix: tuple,
        reduced_stable_items_binding: dict,
        stop_list: list,
        stable_candidates: dict,
        undesired_frame: pd.DataFrame,
        desired_frame: pd.DataFrame,
        new_branches: list,
        verbose: bool,
    ):
        """
        Process stable candidates to generate new branches.

        Parameters
        ----------
        ar_prefix : tuple
            Prefix of the action rule.
        itemset_prefix : tuple
            Prefix of the itemset.
        reduced_stable_items_binding : dict
            Dictionary containing reduced bindings for stable items.
        stop_list : list
            List of stop combinations.
        stable_candidates : dict
            Dictionary containing stable candidates.
        undesired_frame : pd.DataFrame
            Data frame for the undesired state.
        desired_frame : pd.DataFrame
            Data frame for the desired state.
        new_branches : list
            List of new branches generated.
        verbose : bool
            If True, enables verbose output.
        """
        for attribute, items in reduced_stable_items_binding.items():
            for item in items:
                new_ar_prefix = ar_prefix + (item,)
                if self.in_stop_list(new_ar_prefix, stop_list):
                    continue

                undesired_support = undesired_frame[item].sum()
                desired_support = desired_frame[item].sum()

                if verbose:
                    print('SUPPORT')
                    print(itemset_prefix + (item,))
                    print((undesired_support, desired_support))

                if undesired_support < self.min_undesired_support or desired_support < self.min_desired_support:
                    stable_candidates[attribute].remove(item)
                    stop_list.append(new_ar_prefix)
                else:
                    new_branches.append(
                        {
                            'ar_prefix': new_ar_prefix,
                            'itemset_prefix': new_ar_prefix,
                            'item': item,
                            'undesired_mask': undesired_frame[item],
                            'desired_mask': desired_frame[item],
                            'actionable_attributes': 0,
                        }
                    )

    def process_flexible_candidates(
        self,
        ar_prefix: tuple,
        itemset_prefix: tuple,
        reduced_flexible_items_binding: dict,
        stop_list: list,
        stop_list_itemset: list,
        flexible_candidates: dict,
        undesired_frame: pd.DataFrame,
        desired_frame: pd.DataFrame,
        actionable_attributes: int,
        new_branches: list,
        verbose: bool,
    ):
        """
        Process flexible candidates to generate new branches.

        Parameters
        ----------
        ar_prefix : tuple
            Prefix of the action rule.
        itemset_prefix : tuple
            Prefix of the itemset.
        reduced_flexible_items_binding : dict
            Dictionary containing reduced bindings for flexible items.
        stop_list : list
            List of stop combinations.
        stop_list_itemset : list
            List of stop itemsets.
        flexible_candidates : dict
            Dictionary containing flexible candidates.
        undesired_frame : pd.DataFrame
            Data frame for the undesired state.
        desired_frame : pd.DataFrame
            Data frame for the desired state.
        actionable_attributes : int
            Number of actionable attributes.
        new_branches : list
            List of new branches generated.
        verbose : bool
            If True, enables verbose output.
        """
        for attribute, items in reduced_flexible_items_binding.items():
            new_ar_prefix = ar_prefix + (attribute,)
            if self.in_stop_list(new_ar_prefix, stop_list):
                continue

            undesired_states, desired_states, undesired_count, desired_count = self.process_items(
                attribute,
                items,
                itemset_prefix,
                stop_list_itemset,
                undesired_frame,
                desired_frame,
                flexible_candidates,
                verbose,
            )

            if actionable_attributes == 0 and (undesired_count == 0 or desired_count == 0):
                del flexible_candidates[attribute]
                stop_list.append(ar_prefix + (attribute,))
            else:
                for item in items:
                    new_branches.append(
                        {
                            'ar_prefix': new_ar_prefix,
                            'itemset_prefix': itemset_prefix + (item,),
                            'item': item,
                            'undesired_mask': undesired_frame[item],
                            'desired_mask': desired_frame[item],
                            'actionable_attributes': actionable_attributes + 1,
                        }
                    )
                if actionable_attributes + 1 >= self.min_flexible_attributes:
                    self.rules.add_classification_rules(
                        new_ar_prefix,
                        itemset_prefix,
                        undesired_states,
                        desired_states,
                    )

    def process_items(
        self,
        attribute: str,
        items: list,
        itemset_prefix: tuple,
        stop_list_itemset: list,
        undesired_frame: pd.DataFrame,
        desired_frame: pd.DataFrame,
        flexible_candidates: dict,
        verbose: bool,
    ):
        """
        Process items to generate states and counts.

        Parameters
        ----------
        attribute : str
            The attribute being processed.
        items : list
            List of items for the attribute.
        itemset_prefix : tuple
            Prefix of the itemset.
        stop_list_itemset : list
            List of stop itemsets.
        undesired_frame : pd.DataFrame
            Data frame for the undesired state.
        desired_frame : pd.DataFrame
            Data frame for the desired state.
        flexible_candidates : dict
            Dictionary containing flexible candidates.
        verbose : bool
            If True, enables verbose output.

        Returns
        -------
        tuple
            Tuple containing undesired states, desired states, undesired count, and desired count.
        """
        undesired_states = []
        desired_states = []
        undesired_count = 0
        desired_count = 0

        for item in items:
            if self.in_stop_list(itemset_prefix + (item,), stop_list_itemset):
                continue

            undesired_support = undesired_frame[item].sum()
            desired_support = desired_frame[item].sum()

            if verbose:
                print('SUPPORT')
                print(itemset_prefix + (item,))
                print((undesired_support, desired_support))

            undesired_conf = self.rules.calculate_confidence(undesired_support, desired_support)
            if undesired_support >= self.min_undesired_support:
                undesired_count += 1
                if undesired_conf >= self.min_undesired_confidence:
                    undesired_states.append({'item': item, 'support': undesired_support, 'confidence': undesired_conf})

            desired_conf = self.rules.calculate_confidence(desired_support, undesired_support)
            if desired_support >= self.min_desired_support:
                desired_count += 1
                if desired_conf >= self.min_desired_confidence:
                    desired_states.append({'item': item, 'support': desired_support, 'confidence': desired_conf})

            if desired_support < self.min_desired_support and undesired_support < self.min_undesired_support:
                flexible_candidates[attribute].remove(item)
                stop_list_itemset.append(itemset_prefix + (item,))

        return undesired_states, desired_states, undesired_count, desired_count

    def update_new_branches(self, new_branches: list, stable_candidates: dict, flexible_candidates: dict):
        """
        Update new branches with stable and flexible candidates.

        Parameters
        ----------
        new_branches : list
            List of new branches generated.
        stable_candidates : dict
            Dictionary containing stable candidates.
        flexible_candidates : dict
            Dictionary containing flexible candidates.
        """
        for new_branch in new_branches:
            adding = False
            new_stable = {}  # type: dict
            new_flexible = {}  # type: dict

            for attribute, items in stable_candidates.items():
                for item in items:
                    if adding:
                        if attribute not in new_stable:
                            new_stable[attribute] = []
                        new_stable[attribute].append(item)
                    if item == new_branch['item']:
                        adding = True

            for attribute, items in flexible_candidates.items():
                for item in items:
                    if adding:
                        if attribute not in new_flexible:
                            new_flexible[attribute] = []
                        new_flexible[attribute].append(item)
                    if item == new_branch['item']:
                        adding = True

            del new_branch['item']
            new_branch['stable_items_binding'] = new_stable
            new_branch['flexible_items_binding'] = new_flexible

    def in_stop_list(self, ar_prefix, stop_list):
        """
        Check if the action rule prefix is in the stop list.

        Parameters
        ----------
        ar_prefix : tuple
            Prefix of the action rule.
        stop_list : list
            List of stop combinations.

        Returns
        -------
        bool
            True if the action rule prefix is in the stop list, False otherwise.
        """
        if ar_prefix[-2:] in stop_list:
            return True
        if ar_prefix[1:] in stop_list:
            stop_list.append(ar_prefix)
            return True
        return False



#!/usr/bin/env python
"""Tests for `action_rules` package."""

import pytest

from action_rules.output.output import Output


@pytest.fixture
def sample_action_rules():
    """Fixture for sample action rules to be used in tests."""
    return [
        {
            'undesired': {
                'itemset': ['age_<item_stable>_30', 'income_<item_flexible>_low'],
                'support': 10,
                'confidence': 0.8,
                'target': 'status_<item_target>_default',
            },
            'desired': {
                'itemset': ['age_<item_stable>_30', 'income_<item_flexible>_medium'],
                'support': 5,
                'confidence': 0.6,
                'target': 'status_<item_target>_paid',
            },
            'uplift': 0.2,
        }
    ]


@pytest.fixture
def output(sample_action_rules):
    """Fixture for Output instance."""
    return Output(sample_action_rules, 'status')


def test_get_ar_notation(output):
    """Test the get_ar_notation method of Output."""
    ar_notation = output.get_ar_notation()
    assert len(ar_notation) > 0
    assert 'age: 30' in ar_notation[0]


def test_get_export_notation(output):
    """Test the get_export_notation method of Output."""
    export_notation = output.get_export_notation()
    assert len(export_notation) > 0
    assert '"attribute": "age"' in export_notation


def test_get_pretty_ar_notation(output):
    """Test the get_pretty_ar_notation method of Output."""
    pretty_ar_notation = output.get_pretty_ar_notation()
    assert len(pretty_ar_notation) > 0
    assert "If attribute 'age' is '30'" in pretty_ar_notation[0]



#!/usr/bin/env python
"""Tests for `action_rules` package."""

import pytest

from action_rules.rules.rules import Rules


@pytest.fixture
def rules():
    """Fixture for Rules instance."""
    return Rules('status_<item_target>_default', 'status_<item_target>_paid')


def test_add_classification_rules(rules):
    """Test the add_classification_rules method of Rules."""
    new_ar_prefix = tuple()
    itemset_prefix = tuple()
    undesired_states = [{'item': 'age_<item_stable>_30', 'support': 10, 'confidence': 0.8}]
    desired_states = [{'item': 'age_<item_stable>_30', 'support': 5, 'confidence': 0.6}]
    rules.add_classification_rules(new_ar_prefix, itemset_prefix, undesired_states, desired_states)
    assert len(rules.classification_rules[new_ar_prefix]['undesired']) > 0
    assert len(rules.classification_rules[new_ar_prefix]['desired']) > 0


def test_generate_action_rules(rules):
    """Test the generate_action_rules method of Rules."""
    new_ar_prefix = tuple()
    itemset_prefix = tuple()
    undesired_states = [{'item': 'age_<item_stable>_30', 'support': 10, 'confidence': 0.8}]
    desired_states = [{'item': 'age_<item_stable>_30', 'support': 5, 'confidence': 0.6}]
    rules.add_classification_rules(new_ar_prefix, itemset_prefix, undesired_states, desired_states)
    rules.generate_action_rules()
    assert len(rules.action_rules) > 0


def test_prune_classification_rules(rules):
    """Test the prune_classification_rules method of Rules."""
    new_ar_prefix = tuple()
    itemset_prefix = tuple()
    undesired_states = [{'item': 'age_<item_stable>_30', 'support': 10, 'confidence': 0.8}]
    desired_states = []
    rules.add_classification_rules(new_ar_prefix, itemset_prefix, undesired_states, desired_states)
    stop_list = []
    rules.prune_classification_rules(0, stop_list)
    assert len(stop_list) > 0


def test_calculate_confidence(rules):
    """Test the calculate_confidence method of Rules."""
    confidence = rules.calculate_confidence(10, 5)
    assert confidence == 0.6666666666666666


def test_calculate_uplift(rules):
    """Test the calculate_uplift method of Rules."""
    uplift = rules.calculate_uplift(10, 0.8, 0.6)
    assert uplift == 5.0



#!/usr/bin/env python
"""Tests for `action_rules` package."""

import pandas as pd
import pytest

from action_rules.action_rules import ActionRules


@pytest.fixture
def sample_data():
    """Fixture for sample data to be used in tests."""
    data = {
        'age': ['30', '40', '50'],
        'income': ['low', 'medium', 'high'],
        'loan': ['no', 'yes', 'no'],
        'status': ['default', 'paid', 'default'],
    }
    return pd.DataFrame(data)


@pytest.fixture
def action_rules():
    """Fixture for ActionRules instance."""
    return ActionRules(
        min_stable_attributes=1,
        min_flexible_attributes=1,
        min_undesired_support=1,
        min_undesired_confidence=0.5,
        min_desired_support=1,
        min_desired_confidence=0.5,
        verbose=True,
    )


def test_fit(action_rules, sample_data):
    """Test the fit method of ActionRules."""
    action_rules.fit(
        sample_data,
        stable_attributes=['age'],
        flexible_attributes=['income', 'loan'],
        target='status',
        undesired_state='default',
        desired_state='paid',
    )
    assert action_rules.rules is not None
    assert action_rules.output is not None


def test_get_bindings(action_rules, sample_data):
    """Test the get_bindings method of ActionRules."""
    stable_attributes = ['age']
    flexible_attributes = ['income']
    target = 'status'
    data = sample_data.astype(str)
    data_stable = pd.get_dummies(data[stable_attributes], sparse=False, prefix_sep='_<item_stable>_')
    data_flexible = pd.get_dummies(data[flexible_attributes], sparse=False, prefix_sep='_<item_flexible>_')
    data_target = pd.get_dummies(data[[target]], sparse=False, prefix_sep='_<item_target>_')
    data = pd.concat([data_stable, data_flexible, data_target], axis=1)
    stable_items_binding, flexible_items_binding, target_items_binding = action_rules.get_bindings(
        data, stable_attributes, flexible_attributes, target
    )
    assert 'age_<item_stable>_30' in stable_items_binding['age']
    assert 'income_<item_flexible>_low' in flexible_items_binding['income']
    assert 'status_<item_target>_default' in target_items_binding['status']


def test_get_stop_list(action_rules, sample_data):
    """Test the get_stop_list method of ActionRules."""
    stable_attributes = ['age']
    flexible_attributes = ['income']
    target = 'status'
    data = sample_data.astype(str)
    data_stable = pd.get_dummies(data[stable_attributes], sparse=False, prefix_sep='_<item_stable>_')
    data_flexible = pd.get_dummies(data[flexible_attributes], sparse=False, prefix_sep='_<item_flexible>_')
    data_target = pd.get_dummies(data[[target]], sparse=False, prefix_sep='_<item_target>_')
    data = pd.concat([data_stable, data_flexible, data_target], axis=1)
    stable_items_binding, flexible_items_binding, target_items_binding = action_rules.get_bindings(
        data, stable_attributes, flexible_attributes, target
    )
    stop_list = action_rules.get_stop_list(stable_items_binding, flexible_items_binding)
    assert ('age_<item_stable>_30', 'age_<item_stable>_30') in stop_list
    assert ('age_<item_stable>_30', 'age_<item_stable>_40') in stop_list
    assert ('age_<item_stable>_40', 'age_<item_stable>_50') in stop_list
    assert ('income', 'income') in stop_list


def test_get_split_tables(action_rules, sample_data):
    """Test the get_split_tables method of ActionRules."""
    stable_attributes = ['age']
    flexible_attributes = ['income']
    target = 'status'
    data = sample_data.astype(str)
    data_stable = pd.get_dummies(data[stable_attributes], sparse=False, prefix_sep='_<item_stable>_')
    data_flexible = pd.get_dummies(data[flexible_attributes], sparse=False, prefix_sep='_<item_flexible>_')
    data_target = pd.get_dummies(data[[target]], sparse=False, prefix_sep='_<item_target>_')
    data = pd.concat([data_stable, data_flexible, data_target], axis=1)
    _, _, target_items_binding = action_rules.get_bindings(data, stable_attributes, flexible_attributes, target)
    frames = action_rules.get_split_tables(data, target_items_binding, target)
    assert 'status_<item_target>_default' in frames
    assert 'status_<item_target>_paid' in frames


def test_get_rules(action_rules, sample_data):
    """Test the get_rules method of ActionRules."""
    action_rules.fit(
        sample_data,
        stable_attributes=['age'],
        flexible_attributes=['income'],
        target='status',
        undesired_state='default',
        desired_state='paid',
    )
    rules = action_rules.get_rules()
    assert rules is not None



"""Console script for action_rules."""

import os
from typing import BinaryIO

import click
import pandas as pd

from action_rules import ActionRules


@click.command()
@click.option(
    '--min_stable_attributes',
    prompt='Min. stable attributes',
    help='Minimum number of stable attributes.',
    default=1,
    type=int,
)
@click.option(
    '--min_flexible_attributes',
    prompt='Min. flexible attributes',
    help='Minimum number of flexible attributes. Must be at least 1.',
    default=1,
    type=int,
)
@click.option(
    '--min_undesired_support',
    prompt='Min. undesired support',
    help='Support of the undesired part of the rule. Number of instances matching all conditions in the '
    'antecedent and consequent of r1.',
    type=int,
)
@click.option(
    '--min_undesired_confidence',
    prompt='Min. undesired confidence',
    help='Confidence of the undesired part of the rule. Number of instances matching all conditions in the '
    'antecedent and consequent of r1 divided by number of instances matching all conditions in the '
    'antecedent of r1.',
    type=float,
)
@click.option(
    '--min_desired_support',
    prompt='Min. desired support',
    help='Support of the desired part of the rule. Number of instances matching all conditions in the antecedent '
    'and consequent of r2.',
    type=int,
)
@click.option(
    '--min_desired_confidence',
    prompt='Min. desired confidence',
    help='Confidence of the desired part of the rule. Number of instances matching all conditions in the '
    'antecedent and consequent of r2 divided by number of instances matching all conditions in the '
    'antecedent of r2.',
    type=float,
)
@click.option(
    '--csv_path',
    type=click.File('rb'),
    prompt='CSV Path',
    help='Dataset where the first row is the header. A comma is used as a separator.',
)
@click.option(
    '--stable_attributes',
    prompt='Stable attributes (comma separated)',
    help='These attributes remain unchanged regardless of the actions described by the rule.',
    type=str,
)
@click.option(
    '--flexible_attributes',
    prompt='Flexible attributes (comma separated)',
    help='These are the attributes that can change.',
    type=str,
)
@click.option('--target', prompt='Target', help='This is the outcome attribute that the action rule aims to influence.')
@click.option(
    '--undesired_state',
    prompt='Undesired state',
    help='The undesired state of a target is the current or starting state that you want to change or improve. It '
    'represents a negative or less preferred outcome.',
    type=str,
)
@click.option(
    '--desired_state',
    prompt='Desired state',
    help='The desired state of a target is goal state that you want to achieve as a result of applying the action '
    'rule. It represents a positive or preferred outcome.',
    type=str,
)
@click.option(
    '--output_json_path',
    type=click.File('wb'),
    prompt='Output CSV Path',
    help='Action Rules (JSON representation).',
    default='rules.json',
)
def main(
    min_stable_attributes: int,
    min_flexible_attributes: int,
    min_undesired_support: int,
    min_undesired_confidence: float,
    min_desired_support: int,
    min_desired_confidence: float,
    csv_path: BinaryIO,
    stable_attributes: str,
    flexible_attributes: str,
    target: str,
    undesired_state: str,
    desired_state: str,
    output_json_path: BinaryIO,
):
    """
    CLI.

    Decompose a single action rule into two rules, r1 and r2: one representing the state before (undesired part) and
    the other after the intervention (desired part): r1 -> r2.

    Parameters
    ----------
    min_stable_attributes : int
        Minimum number of stable attributes required.
    min_flexible_attributes : int
        Minimum number of flexible attributes required.
    min_undesired_support : int
        Minimum support for the undesired state.
    min_undesired_confidence : float
        Minimum confidence for the undesired state.
    min_desired_support : int
        Minimum support for the desired state.
    min_desired_confidence : float
        Minimum confidence for the desired state.
    csv_path : BinaryIO
        Path to the CSV file containing the dataset.
    stable_attributes : str
        Comma-separated list of stable attributes.
    flexible_attributes : str
        Comma-separated list of flexible attributes.
    target : str
        Target attribute for the action rule.
    undesired_state : str
        The undesired state of the target attribute.
    desired_state : str
        The desired state of the target attribute.
    output_json_path : BinaryIO
        Path to the output JSON file where the results will be saved.

    Returns
    -------
    None
    """
    click.echo("action-rules")
    click.echo("=" * len("action-rules"))
    click.echo("The package for action rules mining using Action-Apriori (Apriori Modified for Action Rules Mining).")
    action_rules = ActionRules(
        int(min_stable_attributes),
        int(min_flexible_attributes),
        int(min_undesired_support),
        float(min_undesired_confidence),
        int(min_desired_support),
        float(min_desired_confidence),
    )
    data = pd.read_csv(os.path.abspath(csv_path.name))
    action_rules.fit(
        data,
        [x.strip() for x in str(stable_attributes).split(",")],
        [x.strip() for x in str(flexible_attributes).split(",")],
        str(target),
        str(undesired_state),
        str(desired_state),
    )
    rules = action_rules.get_rules()
    if rules is not None:
        output_json_path.write(str.encode(str(rules.get_export_notation())))
    click.echo("Done.")


if __name__ == "__main__":
    main()  # pragma: no cover



"""Class Output."""

import json


class Output:
    """
    A class used to format and export action rules.

    Attributes
    ----------
    action_rules : list
        List containing the action rules.
    target : str
        The target attribute for the action rules.

    Methods
    -------
    get_ar_notation()
        Generate a string representation of the action rules in a human-readable format.
    get_export_notation()
        Generate a list of dictionaries representing the action rules for export.
    get_pretty_ar_notation()
        Generate a list of text strings representing the action rules.
    """

    def __init__(self, action_rules: list, target: str):
        """
        Initialize the Output class with the specified action rules and target attribute.

        Parameters
        ----------
        action_rules : list
            List containing the action rules.
        target : str
            The target attribute for the action rules.
        """
        self.action_rules = action_rules
        self.target = target

    def get_ar_notation(self):
        """
        Generate a string representation of the action rules in a human-readable format.

        Returns
        -------
        str
            String representation of the action rules.
        """
        ar_notation = []
        for action_rule in self.action_rules:
            rule = '['
            for i, item in enumerate(action_rule['undesired']['itemset']):
                if i > 0:
                    rule += ' ∧ '
                rule += '('
                if item == action_rule['desired']['itemset'][i]:
                    if '_<item_stable>_' in item:
                        val = item.split('_<item_stable>_')
                        rule += str(val[0]) + ': ' + str(val[1])
                    else:
                        val = item.split('_<item_flexible>_')
                        rule += str(val[0]) + '*: ' + str(val[1])
                else:
                    val = item.split('_<item_flexible>_')
                    val_desired = action_rule['desired']['itemset'][i].split('_<item_flexible>_')
                    rule += str(val[0]) + ': ' + str(val[1]) + ' → ' + str(val_desired[1])
                rule += ')'
            rule += (
                '] ⇒ ['
                + str(self.target)
                + ': '
                + str(action_rule['undesired']['target'])
                + ' → '
                + str(action_rule['desired']['target'])
                + ']'
            )
            rule += (
                ', support of undesired part: '
                + str(action_rule['undesired']['support'])
                + ', confidence of undesired part: '
                + str(action_rule['undesired']['confidence'])
            )
            rule += (
                ', support of desired part: '
                + str(action_rule['desired']['support'])
                + ', confidence of desired part: '
                + str(action_rule['desired']['confidence'])
            )
            rule += ', uplift: ' + str(action_rule['uplift'])
            ar_notation.append(rule)
        return ar_notation

    def get_export_notation(self):
        """
        Generate a list of dictionaries representing the action rules for export.

        Returns
        -------
        list
            List of dictionaries representing the action rules.
        """
        rules = []
        for ar_dict in self.action_rules:
            rule = {'stable': [], 'flexible': []}
            for i, item in enumerate(ar_dict['undesired']['itemset']):
                if item == ar_dict['desired']['itemset'][i]:
                    if '_<item_stable>_' in item:
                        val = item.split('_<item_stable>_')
                        rule['stable'].append({'attribute': val[0], 'value': val[1]})
                    else:
                        val = item.split('_<item_flexible>_')
                        rule['stable'].append({'attribute': val[0], 'value': val[1], 'flexible_as_stable': True})
                else:
                    val = item.split('_<item_flexible>_')
                    val_desired = ar_dict['desired']['itemset'][i].split('_<item_flexible>_')
                    rule['flexible'].append({'attribute': val[0], 'undesired': val[1], 'desired': val_desired[1]})
            rule['target'] = {
                'attribute': self.target,
                'undesired': ar_dict['undesired']['target'].split('_<item_target>_')[1],
                'desired': ar_dict['desired']['target'].split('_<item_target>_')[1],
            }
            rule['support of undesired part'] = int(ar_dict['undesired']['support'])
            rule['confidence of undesired part'] = float(ar_dict['undesired']['confidence'])
            rule['support of desired part'] = int(ar_dict['desired']['support'])
            rule['confidence of desired part'] = float(ar_dict['desired']['confidence'])
            rule['uplift'] = float(ar_dict['uplift'])
            rules.append(rule)
        return json.dumps(rules)

    def get_pretty_ar_notation(self):
        """
        Generate a list of text strings representing the action rules.

        Returns
        -------
        list
            List of text strings representing the action rules.
        """
        rules = []
        for ar_dict in self.action_rules:
            text = "If "
            for i, item in enumerate(ar_dict['undesired']['itemset']):
                if item == ar_dict['desired']['itemset'][i]:
                    if '_<item_stable>_' in item:
                        val = item.split('_<item_stable>_')
                        text += "attribute '" + val[0] + "' is '" + val[1] + "', "
                    else:
                        val = item.split('_<item_flexible>_')
                        text += "attribute (flexible is used as stable) '" + val[0] + "' is '" + val[1] + "', "
                else:
                    val = item.split('_<item_flexible>_')
                    val_desired = ar_dict['desired']['itemset'][i].split('_<item_flexible>_')
                    text += "attribute '" + val[0] + "' value '" + val[1] + "' is changed to '" + val_desired[1] + "', "
            text += (
                "then '"
                + self.target
                + "' value '"
                + ar_dict['undesired']['target']
                + "' is changed to '"
                + ar_dict['desired']['target']
                + " with uplift: "
                + str(ar_dict['uplift'])
                + "."
            )
            rules.append(text)
        return rules



"""Top-level package for Action Rules."""

from .action_rules import ActionRules
from .output.output import Output
from .rules.rules import Rules

__all__ = [
    'ActionRules',
    'Rules',
    'Output',
]
__author__ = """Lukas Sykora"""
__email__ = 'lukas.sykora@vse.cz'
__version__ = '0.0.7'



"""Class Rules."""

from collections import defaultdict


class Rules:
    """
    A class used to manage and generate classification and action rules.

    Attributes
    ----------
    classification_rules : defaultdict
        Default dictionary to store classification rules for undesired and desired states.
    undesired_state : str
        The undesired state of the target attribute.
    desired_state : str
        The desired state of the target attribute.
    action_rules : list
        List to store generated action rules.

    Methods
    -------
    add_classification_rules(new_ar_prefix, itemset_prefix, undesired_states, desired_states)
        Add classification rules for undesired and desired states.
    generate_action_rules()
        Generate action rules from classification rules.
    prune_classification_rules(k, stop_list)
        Prune classification rules based on their length and update the stop list.
    calculate_confidence(support, opposite_support)
        Calculate the confidence of the rule.
    calculate_uplift(undesired_support, undesired_confidence, desired_confidence)
        Calculate the uplift of an action rule.
    """

    def __init__(self, undesired_state: str, desired_state: str):
        """
        Initialize the Rules class with the specified undesired and desired states.

        Parameters
        ----------
        undesired_state : str
            The undesired state of the target attribute.
        desired_state : str
            The desired state of the target attribute.
        """
        self.classification_rules = defaultdict(lambda: {'desired': [], 'undesired': []})  # type: defaultdict
        self.undesired_state = undesired_state
        self.desired_state = desired_state
        self.action_rules = []  # type: list

    def add_classification_rules(self, new_ar_prefix, itemset_prefix, undesired_states, desired_states):
        """
        Add classification rules for undesired and desired states.

        Parameters
        ----------
        new_ar_prefix : tuple
            Prefix of the action rule.
        itemset_prefix : tuple
            Prefix of the itemset.
        undesired_states : list
            List of undesired states.
        desired_states : list
            List of desired states.
        """
        for undesired_item in undesired_states:
            new_itemset_prefix = itemset_prefix + (undesired_item['item'],)
            self.classification_rules[new_ar_prefix]['undesired'].append(
                {
                    'itemset': new_itemset_prefix,
                    'support': undesired_item['support'],
                    'confidence': undesired_item['confidence'],
                    'target': self.undesired_state,
                }
            )
        for desired_item in desired_states:
            new_itemset_prefix = itemset_prefix + (desired_item['item'],)
            self.classification_rules[new_ar_prefix]['desired'].append(
                {
                    'itemset': new_itemset_prefix,
                    'support': desired_item['support'],
                    'confidence': desired_item['confidence'],
                    'target': self.desired_state,
                }
            )

    def generate_action_rules(self):
        """Generate action rules from classification rules."""
        for attribute_prefix, rules in self.classification_rules.items():
            for desired_rule in rules['desired']:
                for undesired_rule in rules['undesired']:
                    uplift = self.calculate_uplift(
                        undesired_rule['support'],
                        undesired_rule['confidence'],
                        desired_rule['confidence'],
                    )
                    self.action_rules.append({'undesired': undesired_rule, 'desired': desired_rule, 'uplift': uplift})

    def prune_classification_rules(self, k: int, stop_list: list):
        """
        Prune classification rules based on their length and update the stop list.

        Parameters
        ----------
        k : int
            Length of the attribute prefix.
        stop_list : list
            List of prefixes to stop generating rules for.
        """
        del_prefixes = []
        for attribute_prefix, rules in self.classification_rules.items():
            if k == len(attribute_prefix):
                if len(rules['desired']) == 0 or len(rules['undesired']) == 0:
                    stop_list.append(attribute_prefix)
                    del_prefixes.append(attribute_prefix)
        for attribute_prefix in del_prefixes:
            del self.classification_rules[attribute_prefix]

    def calculate_confidence(self, support, opposite_support):
        """
        Calculate the confidence of an action rule.

        Parameters
        ----------
        support : int
            The support value for the desired or undesired state.
        opposite_support : int
            The support value for the opposite state.

        Returns
        -------
        float
            The confidence value calculated as support / (support + opposite_support).
            Returns 0 if the sum of support and opposite_support is 0.
        """
        if support + opposite_support == 0:
            return 0
        return support / (support + opposite_support)

    def calculate_uplift(self, undesired_support: int, undesired_confidence: float, desired_confidence: float) -> float:
        """
        Calculate the uplift of an action rule.

        Parameters
        ----------
        undesired_support : int
            The support value for the undesired state.
        undesired_confidence : float
            The confidence value for the undesired state.
        desired_confidence : float
            The confidence value for the desired state.

        Returns
        -------
        float
            The uplift value calculated as:
            (undesired_support / undesired_confidence) * desired_confidence -
            (undesired_support / undesired_confidence - undesired_support).
        """
        return (undesired_support / undesired_confidence) * desired_confidence - (
            undesired_support / undesired_confidence - undesired_support
        )
